import logging
import time
import uuid
import os
import re
import numpy as np
import pandas as pd
from fitxf import FitXformClusterCosine, FitXformCluster, FitXformPca
from fitxf import DbParams
from fitxf.math.lang.encode.LangModelPtSingleton import LangModelPtSingleton
from fitxf.math.lang.encode.LangModelPt import LangModelPt
from fitxf.math.datasource.vecdb.model.ModelDb import ModelDb
from fitxf.math.datasource.vecdb.metadata.Metadata import ModelMetadata
from fitxf import ModelFitTransform, ModelInterface
# from nwae.math.datasource.vecdb.model.ModelStandard import ModelStandard
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env


class VecDbModelUnitTest:

    COL_CONTENT = 'text'
    COL_LABEL_HUMAN = 'label'
    COL_LABEL_NO = 'label_number'
    COL_EMBEDDING = 'embedding'
    COL_CONTENT_TYPE = 'content_type'

    CLUSTER_TYPES = ['cluster', 'cluster-cosine']

    def __init__(
            self,
            user_id: str,
            model_fit_name: str,
            in_plain_or_test_mode: bool,
            logger = None,
    ):
        self.user_id = user_id
        self.model_fit_name = model_fit_name
        self.in_plain_or_test_mode = in_plain_or_test_mode
        self.logger = logger if logger is not None else Logging.get_default_logger(log_level=logging.INFO, propagate=False)

        os.environ['DB_DEFAULT_TIMEOUT'] = '3'
        self.repo_env = Env()
        self.db_params = DbParams.get_db_params_from_envvars(
            identifier = str(self.__class__),
            db_create_tbl_sql = None,
            db_table = self.user_id,
            verify_certs = os.environ["VERIFY_CERTS"].lower() in ['1', 'true', 'yes'],
        )
        self.lm_cache_folder = self.repo_env.MODELS_PRETRAINED_DIR
        self.file_temp_dir = os.environ["TEMP_DIR"]

        self.sample_txt = [
            'hello world', 'cq cq anyone there', 'privet everyone', 'dobriy dyen', '안녕 你好 привет',
            'rainbow in the sky', 'polka dot bikini', 'white nights', 'aurora northern hemisphere',
            'cold and snowing', 'heating at home', 'thick winter jackets', 'gloves and scarfs',
            # Duplicates
            'cq cq anyone there',
            # Test that won't accidentally delete "hello world"
            'world',
        ]
        self.sample_lbl = [
            'hola', 'hola', 'hola', 'hola', 'hola',
            'color', 'color', 'color', 'color',
            'winter', 'winter', 'winter', 'winter',
            # Duplicates & accidental delete tests
            'hola', 'hola',
        ]
        self.sample_records = pd.DataFrame({
            self.COL_CONTENT: self.sample_txt,
            self.COL_CONTENT_TYPE: ModelInterface.TYPE_TEXT,
            self.COL_LABEL_HUMAN: self.sample_lbl,
        }).to_csv("test_data.csv")# to_dict('records')
        raise Exception('asdf')
        self.len_unique = len(np.unique(self.sample_txt))

        self.partial_train_n = [7, len(self.sample_txt)]
        # duplicates are not removed
        self.partial_train_len_trained_no_remove_dup = [7, len(self.sample_txt)]
        self.partial_train_len_trained_hv_remove_dup = [7, self.len_unique]
        logging.info('Setup unit test data done..')
        return

    def test(
            self,
            cleanup = True,
    ):
        vec_dbs_list = self.__test()
        self.logger.info('Tests executed successfully, now cleaning up...')

        if cleanup:
            self.logger.info('Doing final cleanups...')
            for vdb in vec_dbs_list:
                vdb.stop_threads()
                res = vdb.cleanup()
                self.logger.info('Index(es) removed: ' + str(res))
                try:
                    vdb.model_db.get_underlying_db().delete_index(tablename=self.user_id)
                except:
                    pass
        return

    def __get_vecdb_model(self) -> ModelInterface:
        if self.model_fit_name in ['cluster', 'cluster-cosine', 'pca']:
            Class = ModelFitTransform
            if self.model_fit_name == 'cluster':
                fitxfClass = FitXformCluster
            elif self.model_fit_name == 'cluster-cosine':
                fitxfClass = FitXformClusterCosine
            elif self.model_fit_name == 'pca':
                fitxfClass = FitXformPca
            else:
                fitxfClass = None
        else:
            Class = ModelStandard
            fitxfClass = None
        intent = Class(
            user_id = self.user_id,
            llm_model = {
                ModelInterface.TYPE_TEXT: LangModelPtSingleton.get_singleton(
                    LmClass = LangModelPt,
                    cache_folder = self.lm_cache_folder,
                    logger = self.logger,
                ),
            },
            model_db_class = ModelDb,
            model_metadata_class = ModelMetadata,
            col_content = self.COL_CONTENT,
            col_content_type = self.COL_CONTENT_TYPE,
            col_label_user = self.COL_LABEL_HUMAN,
            col_label_std = self.COL_LABEL_NO,
            col_embedding = self.COL_EMBEDDING,
            feature_len = 0,
            numpy_to_b64_for_db = True,
            fit_xform_model = fitxfClass(logger=self.logger) if fitxfClass is not None else None,
            cache_tensor_to_file = True,
            file_temp_dir = self.file_temp_dir,
            in_plain_or_test_mode = self.in_plain_or_test_mode,
            return_tensors = 'np',
            enable_bg_thread_for_training = False,
            logger = self.logger,
        )
        return intent

    def __test(
            self,
    ):
        test_info = 'DATA CACHE "' + str(self.db_params.get_db_info()) \

        self.vec_model = self.__get_vecdb_model()
        assert self.vec_model.get_data_length() == 0, \
            'Data length not zero at start but ' + str(self.vec_model.get_data_length()) \
            + ' for ' + str(self.db_params.get_db_info())

        i_start = 0
        for round, i_end in enumerate(self.partial_train_n):
            self.logger.info(
                'Train round ' + str(round) + ' range start/end [' + str(i_start) + ', ' + str(i_end) + ']'
            )
            # Delete then add, to remove duplicates previously added
            for r in self.sample_records[i_start:i_end]:
                try:
                    self.vec_model.delete(
                        match_phrases = [{self.COL_CONTENT: r[self.COL_CONTENT]}],
                    )
                except:
                    pass
            self.vec_model.add(
                records = self.sample_records[i_start: i_end],
            )
            # call predict to force update model
            self.vec_model.predict(text_list_or_embeddings=['hi'])
            self.logger.info(
                'Partial train round #' + str(round) + ' Current label maps: ' + str(self.vec_model.map_idx_to_lbl)
                + ', ' + str(self.vec_model.map_lbl_to_idx)
            )

            self.logger.info(
                'Map now: ' + str(self.vec_model.map_idx_to_lbl)
                + ', text labels std ' + str(self.vec_model.text_labels_standardized)
            )
            labels_trained = [
                self.vec_model.map_idx_to_lbl[idx] for idx in self.vec_model.text_labels_standardized
            ]
            self.logger.info(
                'Partial train round #' + str(round) + ' Current text labels trained: ' + str(labels_trained)
            )
            assert len(labels_trained) == self.partial_train_len_trained_hv_remove_dup[round],\
                'Partial train round #' + str(round) + ' trained ' + str(len(labels_trained))\
                + ', expected ' + str(self.partial_train_len_trained_hv_remove_dup[round])
            i_start = i_end

        # Test reload new class
        self.vec_model_2 = self.__get_vecdb_model()
        # assert id(tr) == id(tr_2), 'Singleton class must same but different ' + str(id(tr)) + ' vs ' + str(id(tr_2))
        self.logger.info('Loaded new class from old data for db type "' + str(self.db_params.db_type) + '"')
        assert self.vec_model_2.get_data_length() >= self.len_unique,\
            'Model "' + str(self.model_fit_name) + '". After training data cache type "' + str(self.db_params.db_type)\
            + '", length of encoded text from model load ' + str(len(self.vec_model_2.text_labels_standardized))\
            + ' not expected ' + str(self.len_unique)

        do_full_rec_return = type(self.vec_model) not in [ModelFitTransform]
        if do_full_rec_return:
            pred_full_record, pred_probs_full_record = self.vec_model_2.predict(
                text_list_or_embeddings = self.sample_txt,
                return_full_record = True,
            )
        else:
            pred_full_record, pred_probs_full_record = None, None

        pred_labels, pred_probs = self.vec_model_2.predict(
            text_list_or_embeddings = self.sample_txt,
            return_full_record = False,
        )

        score_accuracy = 0.
        len_sample_txt = len(self.sample_txt)
        for i in range(len_sample_txt):
            s = self.sample_txt[i]
            l = self.sample_lbl[i]
            l_pred = pred_labels[i]
            l_prob = pred_probs[i]
            self.logger.info('For line #' + str(i) + ', predictions: ' + str(l_pred))
            msg = '#' + str(i) + ' Compress model "' + str(type(self.vec_model)) + '", for "' + str(s) \
                  + '", expect "' + str(l) + '", got ' + str(l_pred) + ', probs ' + str(l_prob) + ', ' + test_info
            self.logger.debug(msg)
            try:
                top1_user_label_predicted = l_pred[0]['user_label_estimate']
                top2_user_label_predicted = l_pred[1]['user_label_estimate']
            except:
                top1_user_label_predicted = l_pred[0]
                top2_user_label_predicted = l_pred[1]
            if l != top1_user_label_predicted:
                self.logger.warning(msg)
                # if do_assertion:
                #     assert l == l_pred[0], msg
            score_accuracy += min(1.0*(top1_user_label_predicted == l) + 0.5*(top2_user_label_predicted == l), 1.0)

            if do_full_rec_return:
                l_pred_full_rec = pred_full_record[i]
                l_prob_full_rec = pred_probs_full_record[i]
                msg = '#' + str(i) + ' for "' + str(s) \
                      + '", expect "' + str(l) + '", got ' + str(l_pred_full_rec) + ', probs ' + str(l_prob_full_rec) \
                      + ', ' + str(test_info)
                self.logger.info(msg)
                assert l == l_pred_full_rec[0][self.COL_LABEL_HUMAN], msg
        score_accuracy = score_accuracy / len_sample_txt
        acc_thr = 0.8 if self.model_fit_name in ['cluster', 'cluster-cosine'] else 0.9
        self.logger.info('Model Accuracy for "' + str(self.model_fit_name) + '" ' + str(score_accuracy))
        assert score_accuracy > acc_thr, \
            'Model Accuracy for "' + str(self.model_fit_name) + '" ' + str(score_accuracy) + ' < ' + str(acc_thr)

        #
        # Test delete
        #
        labels_before = [
            self.vec_model.map_idx_to_lbl[idx] for idx in self.vec_model.text_labels_standardized
        ]
        self.logger.info('Before delete, labels trained (' + str(len(labels_before)) + '): ' + str(labels_before))
        for mp in [
            {self.COL_CONTENT: 'world'},
            {self.COL_CONTENT: 'world'},
            {self.COL_CONTENT: 'privet everyone'},
        ]:
            self.vec_model.delete(match_phrases = [mp])
        # call predict to force update model
        self.vec_model.predict(text_list_or_embeddings=['hi'])
        self.logger.info(
            'Before calling predict, data records ' + str([r[self.vec_model.col_content] for r in self.vec_model.data_records])
        )
        labels_after = [
            self.vec_model.map_idx_to_lbl[idx] for idx in self.vec_model.text_labels_standardized
        ]
        # self.logger.info('After delete, labels trained (' + str(len(labels_after)) + '):  ' + str(labels_after))
        assert len(labels_before) - len(labels_after) == 2, \
            'Expected in [0, 2] records deleted from before (' + str(len(labels_before)) + ') ' \
            + str(labels_before) + ', after (' + str(len(labels_after)) + ') ' + str(labels_after)

        # Test sync
        if (self.db_params.db_type != 'memory'):
            # sleep to make sure last update time is no longer the same millisecond when calling predict()
            time.sleep(0.1)
            lbls, probs = self.vec_model.predict(
                text_list_or_embeddings = ['is in sync?'],
                # params_other = {'PCA': self.predict_via_model_compression},
            )
            assert lbls[0][0] != 'sync test', 'Result before sync incorrect ' + str([lbls, probs])
            records_test_sync = [
                {self.COL_LABEL_HUMAN: 'sync test', self.COL_LABEL_NO: 999, self.COL_CONTENT_TYPE: ModelInterface.TYPE_TEXT, self.COL_CONTENT: 'I am in sync'},
                {self.COL_LABEL_HUMAN: 'sync test', self.COL_LABEL_NO: 999, self.COL_CONTENT_TYPE: ModelInterface.TYPE_TEXT, self.COL_CONTENT: 'Are you in sync'},
            ]
            # Despite adding from different object, this action will update DB metadata
            self.vec_model_2.add(
                records = records_test_sync,
            )
            records_before_predict = [r[self.vec_model.col_content] for r in self.vec_model.data_records]
            self.logger.info('Before calling predict, data records ' + str(records_before_predict))

            # despite calling from a different object tr, calling predict() will force checking metadata & model update
            lbls, probs = self.vec_model.predict(
                text_list_or_embeddings = ['is in sync?'],
                top_k = 6,
                # params_other = {'PCA': self.predict_via_model_compression},
            )
            self.logger.info('Prediction of sync test: ' + str(lbls) + ', ' + str(probs))
            try:
                top_label_predicted = lbls[0][0]['user_label_estimate']
            except:
                top_label_predicted = lbls[0][0]

            records_after_predict = [r[self.vec_model.col_content] for r in self.vec_model.data_records]
            diff = list(set(records_after_predict).difference(records_before_predict))
            self.logger.info('After calling predict, data records difference ' + str(diff))
            assert top_label_predicted == 'sync test', \
                'For db type "' + str(self.db_params.db_type) + '", result after sync incorrect ' + str([lbls, probs])

        print('Remaining records: ')
        [print(i, r) for i, r in enumerate(self.vec_model.data_records)]

        self.vec_model.check_model_consistency_with_prev()

        self.logger.info(
            'ALL TESTS PASSED OK (' + test_info + ')'
        )
        return [self.vec_model, self.vec_model_2]


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/env.ut.mysql')

    # Put some randomness, so no clashes
    logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False)

    rand_str = re.sub(pattern=".*[\-]", repl="", string=str(uuid.uuid4()))
    bot_unique_id = 'vecdb_model_ut.' + str(rand_str)

    for model, cleanup, plain_mode in [
        # (None, True, False),
        ('pca', True, False),
        ('cluster', True, False),
        ('cluster', True, True),
        ('cluster-cosine', True, False),
    ]:
        for cache_to_file in [False]:
            ut = VecDbModelUnitTest(
                user_id = 'modelut_' + str(model) + '_' + bot_unique_id,
                model_fit_name = model,
                in_plain_or_test_mode = plain_mode,
                logger = logger,
            )
            ut.test(cleanup=cleanup)
            print('TESTS OK FOR cache to file = ' + str(cache_to_file))
    print('ALL TESTS PASSED OK')
    exit(0)



import logging
import time
import uuid
import os
import re
import numpy as np
import pandas as pd
from inspect import getsourcefile
from fitxf import DbParams, ModelInterface
from baimi.datasource.vecdb.VecDb import VecDb
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env

ClassifyClass = VecDb


class VecDbUnitTest:

    COL_CONTENT = 'text'
    COL_CONTENT_TYPE = 'type'
    COL_LABEL_HUMAN = 'label'
    COL_LABEL_NO = 'label_number'
    COL_EMBEDDING = 'embedding'

    CLUSTER_TYPES = ['cluster', 'cluster-cosine']

    def __init__(
            self,
            user_id,
            db_params,
            lm_cache_folder,
            lm_model_name_or_path,
            model_cache_folder,
            is_db_exist,
            cache_tensor_to_file = False,
            file_temp_dir = '/tmp',
            turn_off_fitting = False,
            vdb_model = None,
            logger = None,
    ):
        self.user_id = user_id
        self.db_params = db_params
        self.lm_cache_folder = lm_cache_folder
        self.lm_model_name_or_path = lm_model_name_or_path
        self.model_cache_folder = model_cache_folder
        self.is_db_exist = is_db_exist
        self.vdb_model = vdb_model
        self.cache_tensor_to_file = cache_tensor_to_file
        self.file_temp_dir = file_temp_dir
        self.turn_off_fitting = turn_off_fitting
        self.logger = logger

        os.environ['DB_DEFAULT_TIMEOUT'] = '3'

        self.lang = 'en'
        path = os.path.abspath(getsourcefile(lambda: 0))
        path_folder = os.path.dirname(path)
        self.sample_records_df = pd.read_csv(path_folder + "/test_vecdb.csv")
        self.sample_records = self.sample_records_df.to_dict('records')
        self.sample_txt = list(self.sample_records_df[self.COL_CONTENT])
        self.sample_lbl = list(self.sample_records_df[self.COL_LABEL_HUMAN])

        self.len_unique = len(np.unique(self.sample_records_df[self.COL_CONTENT]))

        self.partial_train_n = [7, len(self.sample_records)]
        # duplicates are not removed
        self.partial_train_len_trained_no_remove_dup = [7, len(self.sample_records)]
        self.partial_train_len_trained_hv_remove_dup = [7, self.len_unique]
        logging.info('Setup unit test data done..')

    def test(
            self,
            use_add_delete = False,
            params_train = None,
            do_assertion = True,
    ):
        vec_dbs_list = self.__test(
            use_add_delete = use_add_delete,
            params_train = params_train,
            do_assertion = do_assertion,
        )
        self.logger.info('Tests executed successfully, now cleaning up...')

        self.logger.info('Doing final cleanups...')
        for vdb in vec_dbs_list:
            res = vdb.cleanup_underlying_db()
            self.logger.info('Index(es) removed: ' + str(res))
            vdb.stop_threads()
        return

    def __get_intent_singleton(self) -> VecDb:
        # For the purpose of test, make it clear memory very often
        os.environ["VECDB_BG_SLEEP_SECS"] = "0.1"
        os.environ["VECDB_CLEAR_MEMORY_SECS_INACTIVE"] = "0.001"
        intent = ClassifyClass(
            user_id = self.user_id,
            lang = self.lang,
            col_content = self.COL_CONTENT,
            col_content_type = self.COL_CONTENT_TYPE,
            col_label_user = self.COL_LABEL_HUMAN,
            col_label_std = self.COL_LABEL_NO,
            col_embedding = self.COL_EMBEDDING,
            feature_len = 768,
            lm_cache_folder = self.lm_cache_folder,
            lm_model_name_or_path = {
                ModelInterface.TYPE_TEXT: self.lm_model_name_or_path,
                ModelInterface.TYPE_IMG: None,
            },
            fit_xform_model_name = self.vdb_model,
            underlying_db_params = self.db_params,
            cache_tensor_to_file = self.cache_tensor_to_file,
            file_temp_dir = self.file_temp_dir,
            turn_off_fitting = self.turn_off_fitting,
            logger = self.logger,
        )
        return intent

    def __test(
            self,
            use_add_delete = False,
            params_train = None,
            do_assertion = True,
    ):
        params_train = {} if params_train is None else params_train

        test_info = 'DATA CACHE "' + str(self.db_params.get_db_info()) \
                    + '", ADD/DELETE INSTEAD OF TRAIN=' + str(use_add_delete) + ')'

        ClassifyClass.MODEL_COMPRESSION_SLEEP_SECS = 0.5

        tr = self.__get_intent_singleton()
        self.vec_db = tr
        tr.cleanup_underlying_db()
        self.logger.info('Model lengths after clearing DB at start = ' + str(tr.get_data_length()))
        assert tr.get_data_length() == 0, \
                'Data length not zero at start but ' + str(tr.get_data_length()) + ' for ' + str(self.db_params.get_db_info())

        i_start = 0
        for round, i_end in enumerate(self.partial_train_n):
            self.logger.info(
                'Train round ' + str(round) + ' range start/end [' + str(i_start) + ', ' + str(i_end)
                + '], params train: ' + str(params_train) + ', use add/delete ' + str(use_add_delete)
            )
            # Delete then add, to remove duplicates previously added
            for r in self.sample_records[i_start:i_end]:
                # Only do delete after first round, so that table/etc already created
                if round > 0:
                    tr.delete(
                        match_phrase = [{self.COL_CONTENT: r[self.COL_CONTENT]}],
                    )
            tr.add(
                records = self.sample_records[i_start: i_end],
                model = None,
            )
            # call predict to force update model
            tr.predict(text_list_or_embeddings=['hi'])
            self.logger.info(
                'Partial train round #' + str(round) + ' Current label maps: ' + str(tr.vec_db_model.map_idx_to_lbl)
                + ', ' + str(tr.vec_db_model.map_lbl_to_idx)
            )

            self.logger.info(
                'Map now: ' + str(tr.vec_db_model.map_idx_to_lbl)
                + ', text labels std ' + str(tr.vec_db_model.text_labels_standardized)
            )
            labels_trained = [
                tr.vec_db_model.map_idx_to_lbl[idx] for idx in tr.vec_db_model.text_labels_standardized
            ]
            self.logger.info(
                'Partial train round #' + str(round) + ' Current text labels trained: ' + str(labels_trained)
            )
            assert len(labels_trained) == self.partial_train_len_trained_hv_remove_dup[round],\
                'Partial train round #' + str(round) + ' trained ' + str(len(labels_trained))\
                + ', expected ' + str(self.partial_train_len_trained_hv_remove_dup[round])
            i_start = i_end

        if (self.db_params.db_type == 'memory'):
            # For memory type, without model files, it will have nothing if instantiate new class, so cannot do that
            tr_2 = tr
        else:
            # Test reload new class
            tr_2 = self.__get_intent_singleton()
            # assert id(tr) == id(tr_2), 'Singleton class must same but different ' + str(id(tr)) + ' vs ' + str(id(tr_2))
            self.logger.info('Loaded new class from old data for db type "' + str(self.db_params.db_type) + '"')
        self.vec_db_2 = tr_2
        assert tr_2.get_data_length() >= self.len_unique,\
            'After training data cache type "' + str(db_params.db_type)\
            + '", length of encoded text from model load ' + str(len(tr_2.vec_db_model.text_labels_standardized))\
            + ' not expected ' + str(self.len_unique)

        do_full_rec_return = self.vdb_model not in self.CLUSTER_TYPES
        if do_full_rec_return:
            pred_full_record, pred_probs_full_record = tr_2.predict(
                text_list_or_embeddings = self.sample_txt,
                return_full_record = True,
            )
        else:
            pred_full_record, pred_probs_full_record = None, None

        pred_labels, pred_probs = tr_2.predict(
            text_list_or_embeddings = self.sample_txt,
            return_full_record = False,
        )

        score_accuracy = 0.
        len_sample_txt = len(self.sample_txt)
        for i in range(len_sample_txt):
            s = self.sample_txt[i]
            l = self.sample_lbl[i]
            l_pred = pred_labels[i]
            l_prob = pred_probs[i]
            self.logger.info('For line #' + str(i) + ', predictions: ' + str(l_pred))
            msg = '#' + str(i) + ' Compress model "' + str(self.vdb_model) + '", for "' + str(s) \
                  + '", expect "' + str(l) + '", got ' + str(l_pred) + ', probs ' + str(l_prob) + ', ' + test_info
            self.logger.info(msg)
            top1_user_label_predicted \
                = l_pred[0]['user_label_estimate'] if self.vdb_model in self.CLUSTER_TYPES else l_pred[0]
            top2_user_label_predicted \
                = l_pred[1]['user_label_estimate'] if self.vdb_model in self.CLUSTER_TYPES else l_pred[1]
            if l != top1_user_label_predicted:
                self.logger.warning(msg)
                # if do_assertion:
                #     assert l == l_pred[0], msg
            score_accuracy += min(1.0*(top1_user_label_predicted == l) + 0.5*(top2_user_label_predicted == l), 1.0)

            if do_full_rec_return:
                l_pred_full_rec = pred_full_record[i]
                l_prob_full_rec = pred_probs_full_record[i]
                msg = '#' + str(i) + ' Compress model "' + str(self.vdb_model) + '", for "' + str(s) \
                      + '", expect "' + str(l) + '", got ' + str(l_pred_full_rec) + ', probs ' + str(l_prob_full_rec) \
                      + ', ' + str(test_info)
                self.logger.info(msg)
                # raise Exception('asdf')
                if do_assertion:
                    assert l == l_pred_full_rec[0][self.COL_LABEL_HUMAN], msg
        score_accuracy = score_accuracy / len_sample_txt
        acc_thr = 0.85
        assert score_accuracy > acc_thr, \
            'Compress model "' + str(self.vdb_model) + ' accuracy ' + str(score_accuracy) + ' < ' + str(acc_thr)

        #
        # Test delete
        #
        labels_before = [
            tr.vec_db_model.map_idx_to_lbl[idx] for idx in tr.vec_db_model.text_labels_standardized
        ]
        self.logger.info('Before delete, labels trained (' + str(len(labels_before)) + '): ' + str(labels_before))
        for mp in [
            {self.COL_CONTENT: 'world'},
            {self.COL_CONTENT: 'world'},
            {self.COL_CONTENT: 'privet everyone'},
            {self.COL_CONTENT_TYPE: ModelInterface.TYPE_IMG, self.COL_CONTENT: 'https://assets.teenvogue.com/photos/66f2dbe59926099a9dec34ce/16:9/w_2240,c_limit/2173404736'},
        ]:
            try:
                tr.delete(match_phrase = [mp])
            except Exception as ex_del:
                self.logger.info('Delete error for match phrase ' + str(mp) + ': ' + str(ex_del))
        # call predict to force update model
        tr.predict(text_list_or_embeddings=['hi'])
        self.logger.info(
            'Before calling predict, data records ' + str([r[tr_2.col_content] for r in tr_2.vec_db_model.data_records])
        )
        labels_after = [
            tr.vec_db_model.map_idx_to_lbl[idx] for idx in tr.vec_db_model.text_labels_standardized
        ]
        # self.logger.info('After delete, labels trained (' + str(len(labels_after)) + '):  ' + str(labels_after))
        assert len(labels_before) - len(labels_after) == 2, \
            'Expected in [0, 2] records deleted from before (' + str(len(labels_before)) + ') ' \
            + str(labels_before) + ', after (' + str(len(labels_after)) + ') ' + str(labels_after)

        # Test sync
        if (self.db_params.db_type != 'memory'):
            # sleep to make sure last update time is no longer the same millisecond when calling predict()
            time.sleep(0.1)
            lbls, probs = tr.predict(
                text_list_or_embeddings = ['is in sync?'],
                # params_other = {'PCA': self.predict_via_model_compression},
            )
            assert lbls[0][0] != 'sync test', 'Result before sync incorrect ' + str([lbls, probs])
            records_test_sync = [
                {self.COL_LABEL_HUMAN: 'sync test', self.COL_LABEL_NO: 999, self.COL_CONTENT_TYPE: 'text', self.COL_CONTENT: 'I am in sync'},
                {self.COL_LABEL_HUMAN: 'sync test', self.COL_LABEL_NO: 999, self.COL_CONTENT_TYPE: 'text', self.COL_CONTENT: 'Are you in sync'},
            ]
            # Despite adding from different object, this action will update DB metadata
            tr_2.add(
                records = records_test_sync,
            )
            records_before_predict = [r[tr_2.col_content] for r in tr.vec_db_model.data_records]
            self.logger.info('Before calling predict, data records ' + str(records_before_predict))

            # despite calling from a different object tr, calling predict() will force checking metadata & model update
            lbls, probs = tr_2.predict(
                text_list_or_embeddings = ['is in sync?'],
                top_k = 6,
                # params_other = {'PCA': self.predict_via_model_compression},
            )
            self.logger.info('Prediction of sync test: ' + str(lbls) + ', ' + str(probs))
            top_label_predicted = \
                lbls[0][0]['user_label_estimate'] if self.vdb_model in self.CLUSTER_TYPES else lbls[0][0]

            records_after_predict = [r[tr_2.col_content] for r in tr.vec_db_model.data_records]
            diff = list(set(records_after_predict).difference(records_before_predict))
            self.logger.info('After calling predict, data records difference ' + str(diff))
            assert top_label_predicted == 'sync test', \
                'For db type "' + str(self.db_params.db_type) + '", model predict = "' \
                + str(self.vdb_model) + '", result after sync incorrect ' + str([lbls, probs])

        print('Remaining records: ')
        [print(i, r) for i, r in enumerate(tr.vec_db_model.data_records)]

        tr.vec_db_model.check_model_consistency_with_prev()

        self.logger.info(
            'ALL TESTS PASSED OK (' + test_info + ')'
        )
        return [tr, tr_2]


if __name__ == '__main__':
    UnitTestClass = VecDbUnitTest
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = Env(logger=lgr)
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/env.ut.mysql')

    rand_str = re.sub(pattern=".*[\-]", repl="", string=str(uuid.uuid4()))
    bot_unique_id = 'bot_intent_unit_test_data_index.' + str(rand_str)

    cache_tensor_to_file = True
    db_params = DbParams(
        identifier = str(UnitTestClass),
        logger = lgr,
    )

    ut = UnitTestClass(
        user_id = bot_unique_id,
        db_params = db_params,
        lm_cache_folder = er.MODELS_PRETRAINED_DIR,
        lm_model_name_or_path = None,
        model_cache_folder = None,
        is_db_exist = True,
        vdb_model = os.environ["VECDB_FIT_XFORM_MODEL"],
        cache_tensor_to_file = cache_tensor_to_file,
        file_temp_dir = ".",
        turn_off_fitting = str(os.environ["VECDB_FIT_XFORM_MODEL_TEST_MODE"]).lower() in ['1', 'yes', 'true'],
        logger = lgr,
    )
    ut.test()
    print('TESTS OK FOR cache to file = ' + str(cache_tensor_to_file))
    print('ALL TESTS PASSED OK')
    exit(0)



import os
import re
import time
import numpy as np
from fitxf.math.algo.encoding.Base64 import Base64
from fitxf import ModelDbInterface
from fitxf.math.datasource.DbLang import DbLang
from fitxf import DbParams, DatastoreInterface, Datastore
from fitxf.math.utils.Lock import Lock


class ModelDb(ModelDbInterface):

    def __init__(
            self,
            tablename: str,
            col_content: str,
            col_content_type: str,
            col_label_user: str,
            col_label_standardized: str,
            col_embedding: str,
            max_records: int,
            logger = None,
    ):
        super().__init__(
            tablename = tablename,
            col_content = col_content,
            col_content_type = col_content_type,
            col_label_user = col_label_user,
            col_label_standardized = col_label_standardized,
            col_embedding = col_embedding,
            max_records = max_records,
            logger = logger,
        )
        self.base64_encoder = Base64(logger=self.logger)
        self.__mutex_data = 'underlying_db_mutex'
        self.__lock = Lock(
            mutex_names = [self.__mutex_data],
            logger = self.logger,
        )
        self.connect_to_underlying_db()
        return

    def get_create_table_db_cmd(
            self,
            db_type,
    ) -> str:
        template = DbLang(logger=self.logger).get_db_syntax_create_table(
            db_type = db_type,
            tablename = "<TABLENAME>",
            columns = [
                "`" + str(self.col_content) + "` TEXT NOT NULL",
                "`" + str(self.col_content_type) + "` varchar(32) NOT NULL DEFAULT 'text'",
                "`" + str(self.col_label_user) + "` varchar(255) NOT NULL",
                "`" + str(self.col_label_standardized) + "` int NOT NULL",
                "`" + str(self.col_embedding) + "` TEXT NOT NULL",
            ],
        )
        self.logger.info('DB create table template for "' + str(db_type) + '": ' + str(template))
        if template is None:
            return None
        return template

    def get_db_params(self) -> DbParams:
        return self.db_params

    def get_underlying_db(self) -> DatastoreInterface:
        return self.underlying_db

    def connect_to_underlying_db(
            self,
    ):
        self.underlying_db = Datastore(
            db_params = self.db_params,
            logger = self.logger,
        ).get_data_store()
        self.logger.info('Connected to underlying DB ' + str(self.db_params.get_db_info()))
        return

    def insert(
            self,
            records: list[dict],
            tablename: str,
    ):
        return self.underlying_db.add(
            records = records,
            tablename = tablename,
        )

    def delete(
            self,
            match_phrase: dict,
            tablename: str,
    ):
        return self.underlying_db.delete(
            match_phrase = match_phrase,
            tablename = tablename,
        )

    def load_data(
            self,
            max_attemps = 1,
    ):
        try:
            self.__lock.acquire_mutexes(
                id = 'load_data',
                mutexes = [self.__mutex_data],
            )
            return self.__load_data(max_attempts=max_attemps)
        finally:
            self.__lock.release_mutexes(mutexes=[self.__mutex_data])

    def __load_data(
            self,
            max_attempts = 1,
    ):
        try_count = 0
        while True:
            try_count += 1
            try:
                records_from_db = self.__load_data_job()
                self.logger.info(
                    'Done load from underlying DB, loaded data len ' + str(len(records_from_db))
                    + ', try no. ' + str(try_count) + '.'
                )
                # if try_count < 3:
                #     raise Exception('Fake fail count ' + str(try_count))
                return records_from_db
            except Exception as ex:
                self.logger.error('Loading from underlying DB failed ' + str(try_count) + ' times: ' + str(ex))
                if try_count >= max_attempts:
                    raise Exception('Loading from primary DB max tries reached ' + str(max_attempts))
                # Sleep for some random time 2-5secs
                time.sleep(2+3*np.random.rand())

    #
    # This critical function keeps our model updated with whatever in DB/memory cache
    #
    def __load_data_job(
            self,
    ):
        try:
            # Get all encoded texts, standardized labels, text encoding from memory cache
            all_records = self.underlying_db.get_all(
                key = self.col_content,
                tablename = self.tablename,
                max_records = self.max_records,
                request_timeout = int(os.environ.get('DATA_LOAD_REQUEST_TIMEOUT', 30)),
            )
            self.logger.info('Retrieved all records total length ' + str(len(all_records)))
            # if not all_records:
            #     raise Exception(
            #         'No data retrieved from data cache "' + str(self.db_params.db_info) + '"'
            #     )

            # Embedding column will be removed below
            return all_records
        except Exception as ex:
            errmsg = 'Error loading model from underlying DB' + str(self.db_params.get_db_info()) \
                     + ' Exception message: ' + str(ex)
            self.logger.error(errmsg)
            raise Exception(errmsg)


    def convert_csv_string_array_to_float_array(
            self,
            string_array,
            custom_chars_remove,
    ):
        return self.underlying_db.convert_csv_string_array_to_float_array(
            string_array = string_array,
            custom_chars_remove = custom_chars_remove,
        )


if __name__ == '__main__':
    exit(0)



import json
import logging
import os
import threading
import numpy as np
from datetime import datetime
from fitxf import MetadataInterface
from fitxf import DbParams, Datastore, DbLang
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Env import Env


class ModelMetadata(MetadataInterface):

    def __init__(
            self,
            # name to identify which user/table/etc this metadata is referring to
            user_id,
            metadata_tbl_name = 'model_metadata',
            logger = None,
    ):
        super().__init__(
            user_id = user_id,
            metadata_tbl_name = metadata_tbl_name,
            logger = logger,
        )

        self.db_params_metadata = DbParams.get_db_params_from_envvars(
            identifier = str(self.__class__),
            # fill in later below
            db_create_tbl_sql = None,
            db_table = self.metadata_tbl_name,
            verify_certs = os.environ.get("VERIFY_CERTS", "1").lower() in ['1', 'true', 'yes'],
        )

        # For metadata
        # Change create table syntax
        if self.db_params_metadata.db_type in ('mysql',):
            self.db_params_metadata.db_create_table_sql = DbLang.get_db_syntax_create_table_mysql(
                tablename = "<TABLENAME>",
                columns = [
                    "`" + str(self.COL_METADATA_USERID) + "` varchar(255) NOT NULL",
                    "`" + str(self.COL_METADATA_IDENTIFIER) + "` varchar(255) NOT NULL",
                    "`" + str(self.COL_METADATA_TIMESTAMP) + "` double DEFAULT NULL",
                    "`" + str(self.COL_METADATA_VALUE) + "` varchar(5000) DEFAULT NULL",
                ],
            )
        # + ', PRIMARY KEY (' + str(self.COL_METADATA_INDEX) + ', ' + str(self.COL_METADATA_IDENTIFIER) + ')' \
        self.logger.info(
            'Using DB create table sql syntax as "' + str(self.db_params_metadata.db_create_table_sql)
            + '" for DB type "' + str(self.db_params_metadata.db_type) + '"'
        )

        # Model params, last update times, etc.
        self.db_metadata = Datastore(
            db_params = self.db_params_metadata,
            logger = self.logger,
        ).get_data_store()
        self.logger.info('Connected to underlying metadata DB ' + str(self.db_params_metadata.get_db_info()))

        self.__mutex_db = threading.Lock()
        self.last_cleanup_time = datetime(year=2000, month=1, day=1)
        self.min_interval_secs_cleanup = 30
        return

    def __get_specific_metadata(
            self,
            metadata_identifier,
    ):
        try:
            self.__mutex_db.acquire()

            match_phrase = {
                self.COL_METADATA_USERID: self.user_id,
                self.COL_METADATA_IDENTIFIER: metadata_identifier,
            }
            rows = self.db_metadata.get(
                match_phrase = match_phrase,
                tablename = self.db_params_metadata.db_table,
            )
            if len(rows) == 0:
                raise Exception('No metadata returned for ' + str(match_phrase) + '. Returned rows ' + str(rows))
            elif len(rows) == 1:
                return rows[0]
            else:
                last_timestamp = -np.inf
                row_keep = None
                for r in rows:
                    timestamp = r[self.COL_METADATA_TIMESTAMP]
                    if timestamp > last_timestamp:
                        row_keep = r

                self.logger.debug(
                    'Metadata returned > 1 rows from table/index "' + str(self.db_params_metadata.db_table)
                    + '": ' + str(rows) + ', keep: ' + str(row_keep)
                )
                return row_keep
        except Exception as ex:
            self.logger.error(
                'Error getting metadata "' + str(metadata_identifier) + '", returning "' + str(None) + '": ' + str(ex)
            )
        finally:
            self.__mutex_db.release()

    def get_metadata(
            self,
            identifier,
    ):
        try:
            row = self.__get_specific_metadata(
                metadata_identifier = identifier,
            )
            value = row[self.COL_METADATA_VALUE]
            try:
                row[self.COL_METADATA_VALUE] = json.loads(value)
            except:
                pass
            return row
        except Exception as ex:
            self.logger.error('Error get metadata identifier "' + str(identifier) + '": ' + str(ex))
            return None

    # signify that model has been updated
    def update_metadata_identifier_value(
            self,
            identifier: str,
            value: str,
    ):
        if type(value) is dict:
            self.logger.warning(
                'Value for metadata identifier "' + str(identifier)
                + '" is not str but dict. Will try to convert value: ' + str(value)
            )
            try:
                value_str = json.dumps(value)
            except Exception as ex:
                self.logger.error(
                    'Failed to convert value for metadata identifier "' + str(identifier) + '", value ' + str(value)
                    + ', exception: ' + str(ex)
                )
                value_str = str(value)
        else:
            value_str = value
        insert_records = [
            {
                self.COL_METADATA_USERID: self.user_id,
                self.COL_METADATA_IDENTIFIER: identifier,
                self.COL_METADATA_TIMESTAMP: self.get_timestamp(),
                self.COL_METADATA_VALUE: value_str,
            }
        ]
        self.logger.debug('Try to update metadata with records: ' + str(insert_records))
        return self.__update_metadata_to_db(
            tablename = self.db_params_metadata.db_table,
            records = insert_records,
        )

    def __update_metadata_to_db(
            self,
            tablename,
            records,
    ):
        # add timestamp to records
        for r in records:
            r[self.COL_METADATA_TIMESTAMP] = self.get_timestamp()

        try:
            self.__mutex_db.acquire()
            tdif_cleanup = datetime.now() - self.last_cleanup_time
            tdif_cleanup_secs = tdif_cleanup.days * 86400 + tdif_cleanup.seconds + tdif_cleanup.microseconds / 1000000
            if tdif_cleanup_secs > self.min_interval_secs_cleanup:
                for mp in [
                    {
                        self.COL_METADATA_USERID: d[self.COL_METADATA_USERID],
                        self.COL_METADATA_IDENTIFIER: d[self.COL_METADATA_IDENTIFIER]
                    } for d in records
                ]:
                    try:
                        # Table might not exist if first time
                        res = self.db_metadata.delete(
                            tablename = tablename,
                            match_phrase = mp,
                        )
                        self.logger.info(
                            'Deleted from metadata table using match phrase ' + str(mp) + ', result ' + str(res)
                        )
                    except Exception as ex:
                        # For
                        self.logger.error('Error deleting metadata with match phrase "' + str(mp) + '": ' + str(ex))
                    # update last cleanup time
                    self.last_cleanup_time = datetime.now()
            else:
                self.logger.info('Ignore metadata cleanup, last done only ' + str(tdif_cleanup_secs) + ' ago')

            self.db_metadata.add(
                tablename = tablename,
                records = records,
            )
            self.logger.info(
                'Successfully wrote metadata records to "' + str(tablename) + '": '
                + str([{k: str(v)[0:min(300,len(str(v)))] for k, v in r.items()} for r in records])
            )
            return records
        finally:
            self.__mutex_db.release()

    def cleanup(
            self,
    ):
        try:
            self.__mutex_db.acquire()
            res = self.db_metadata.delete(
                match_phrase = {self.COL_METADATA_USERID: self.user_id},
                tablename = self.db_params_metadata.db_table,
            )
            self.logger.info('Successfully deleted metadata for index "' + str(self.user_id) + '": ' + str(res))
            return {'deleted': res['deleted']}
        except Exception as ex:
            self.logger.info('Error delete metadata for index "' + str(self.user_id) + '": ' + str(ex))
        finally:
            self.__mutex_db.release()


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut.mysql')
    md = ModelMetadata(
        user_id = 'test',
        logger = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    )
    for id, val in [('test', '123'), ('test_dict', {'name': 'jane', 'age': 55})]:
        res = md.update_metadata_identifier_value(
            identifier = id,
            value = val,
        )
        print('Wrote id ' + str(id) + ', val ' + str(val) + ', res: ' + str(res))
    exit(0)



import threading
import logging
import os
import pandas as pd
from fitxf import DbParams
from baimi.datasource.vecdb.VecDb import VecDb
from fitxf.math.utils.Logging import Logging
from fitxf.math.utils.Pandas import Pandas
from fitxf.math.utils.Env import Env


class VecDbCcrcyTest:

    def __init__(
            self,
            test_data_csv_path: str,
            max_records: int,
            user_id: str,
            col_content_type: str,
            col_label: str,
            col_label_number: str,
            col_text: str,
            feature_len: int,
            lm_cache_folder: str,
            logger = None,
    ):
        self.test_data_csv_path = test_data_csv_path
        self.max_records = max_records
        self.user_id = user_id
        self.col_content_type = col_content_type
        self.col_label = col_label
        self.col_label_number = col_label_number
        self.col_text = col_text
        self.col_embedding = 'embedding'
        self.feature_len = feature_len
        self.lm_cache_folder = lm_cache_folder
        self.logger = logger if logger is not None else logging.getLogger()

        self.db_params = DbParams(identifier=str(self.__class__), logger=self.logger)

        # For the purpose of test, make it clear memory very often
        os.environ["VECDB_BG_SLEEP_SECS"] = "1."
        os.environ["VECDB_CLEAR_MEMORY_SECS_INACTIVE"] = "0.5"
        self.vecdb = VecDb(
            user_id = self.user_id,
            lang = 'en',
            col_content = self.col_text,
            col_content_type = self.col_content_type,
            col_label_user = self.col_label,
            col_label_std = self.col_label_number,
            col_embedding = self.col_embedding,
            feature_len = self.feature_len,
            lm_cache_folder = self.lm_cache_folder,
            underlying_db_params = self.db_params,
            fit_xform_model_name = os.environ["VECDB_FIT_XFORM_MODEL"],
            logger = self.logger,
        )
        # clear table from previous tests, if any
        self.vecdb.delete_index(tablename_or_index=user_id)

        Pandas.increase_display()
        df = pd.read_csv(filepath_or_buffer=self.test_data_csv_path, sep=',', index_col=False)
        df[self.col_content_type] = 'text'
        columns_keep = [self.col_content_type, self.col_label, self.col_text]
        df = df[columns_keep]
        df.dropna(inplace=True)
        if len(df) > self.max_records:
            df = df[0:max_records].reset_index(drop=True)

        self.df = df
        # _, _, df[self.col_label_std] = FitUtils().map_labels_to_consecutive_numbers(lbl_list=list(df[self.col_label]))
        self.records = df.to_dict('records')
        self.logger.info('Successfully read data of shape ' + str(self.df.shape))
        self.unique_texts = list(pd.unique(df[self.col_text]))
        self.unique_texts.sort()
        self.logger.info('Unique texts count ' + str(len(self.unique_texts)))
        self.test_stats = {'predict': {}, 'add/delete': {}}
        self.__mutex = threading.Lock()
        # [print(i, txt) for i, txt in enumerate(self.unique_texts)]
        return

    def update_test_stats(self, thread_name, predict_ok=0, add_del_ok=0):
        try:
            self.__mutex.acquire()
            if thread_name not in self.test_stats['predict'].keys():
                self.test_stats['predict'][thread_name] = 0
            if thread_name not in self.test_stats['add/delete'].keys():
                self.test_stats['add/delete'][thread_name] = 0
            self.test_stats['predict'][thread_name] += predict_ok
            self.test_stats['add/delete'][thread_name] += add_del_ok
        finally:
            self.__mutex.release()

    def get_thread_name(self, i):
        thr_name = 't#' + str(i)
        return thr_name

    def test(
            self,
            n_threads: int,
    ):
        threads = {}
        for i in range(n_threads):
            thr_name = self.get_thread_name(i=i)
            threads[i] = threading.Thread(
                target = self.thread_test,
                name = thr_name,
                args = [thr_name],
            )
            threads[i].start()
        for i in range(n_threads):
            threads[i].join()
            self.logger.info('Thread "' + str(self.get_thread_name(i=i)) + '" finished')

        records = self.vecdb.get_all(key=self.col_label, tablename_or_index=self.user_id)
        res_no_embed = [{k:v for k,v in r.items() if k not in [self.col_embedding]} for r in records]
        final_text_list = [r[self.col_text] for r in records]
        df_final = pd.DataFrame(res_no_embed).sort_values(by=[self.col_text], ascending=True)
        print(df_final)
        self.vecdb.stop_threads()

        # Check if all unique inside
        for txt in self.unique_texts:
            assert txt in final_text_list, 'Text "' + str(txt) + '" not in final list of texts ' + str(final_text_list)
        assert len(res_no_embed) == len(self.unique_texts), \
                'Final records count ' + str(len(res_no_embed)) \
                + ' not equal to unique texts ' + str(len(self.unique_texts))
        self.logger.info('VEC DB CONCURRENCY TESTS PASSED')
        print('VEC DB CONCURRENCY TESTS PASSED')
        return

    def thread_test(
            self,
            name: str,
    ):
        self.logger.info('Start thread "' + str(name) + '", records total ' + str(len(self.records)))
        for i, rec in enumerate(self.records):
            # sleep a little
            # time.sleep(np.random.rand()*0.1)
            self.logger.info(
                'Thread "' + str(name) + '" sending record #' + str(i) + ': '
                + str({k:v for k,v in rec.items() if k != self.col_embedding})
            )
            self.logger.info('Thread "' + str(name) + '" sending request #' + str(i) + ': ' + str(rec))
            _ = self.vecdb.atomic_delete_add(
                delete_key = self.col_text,
                records = [rec],
                model = None,
            )
            self.update_test_stats(thread_name=name, add_del_ok=1)
            # _ = self.vecdb.add(records=[rec])
            try:
                top_lbls, _ = self.vecdb.predict(
                    text_list_or_embeddings = [rec[self.col_text]],
                    model = None,
                )
                assert top_lbls[0][0] == rec[self.col_label], \
                    'Predicted top label "' + str(top_lbls[0][0]) + '" not "' + str(rec[self.col_label]) \
                    + '", for text "' + str(rec[self.col_text]) + '"'
                # res_no_embed = [{k: v for k, v in r.items() if k not in [self.col_embedding]} for r in response]
                # self.logger.info('Thread "' + str(name) + '" response #' + str(i) + ': ' + str(response))
                self.update_test_stats(thread_name=name, predict_ok=1)
            except Exception as ex:
                self.logger.error('Error test predict: ' + str(ex))
        return


if __name__ == '__main__':
    evrepo = Env()
    Env.set_env_vars_from_file(env_filepath=evrepo.REPO_DIR + '/env.ut.mysql')
    csv_path = evrepo.NLP_DATASET_DIR + '/lang-model-test/data.csv'
    lgr = Logging.get_logger_from_env_var()
    bot_id = 'bot_vecdb_ccrcy_test'
    col_cont_type = 'type'
    col_label = 'label'
    col_label_number = '__label'
    col_text = 'text'

    VecDbCcrcyTest(
        test_data_csv_path = csv_path,
        max_records = 50,
        user_id = bot_id,
        col_content_type = col_cont_type,
        col_label = col_label,
        col_label_number = col_label_number,
        col_text = col_text,
        feature_len = 384,
        lm_cache_folder = evrepo.MODELS_PRETRAINED_DIR,
        logger = lgr,
    ).test(
        n_threads = 10,
    )
    exit(0)



import logging
import os
from fitxf import DbParams, ModelInterface
from baimi.datasource.vecdb.VecDb import VecDb
from fitxf.math.utils.Singleton import Singleton
from fitxf.math.utils.Env import Env


class VecDbSingleton:

    @staticmethod
    def get_singleton(
            VecDbClass,    # class type
            user_id: str,
            lang: str,
            col_content: str,
            # text, image, sound, video, voice, etc.
            col_content_type: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            feature_len: int,
            underlying_db_params: DbParams,
            # Where we cache the language models
            lm_cache_folder: str = None,
            # Language model name, e.g. "xlm-roberta-base", if empty will use default map by language
            lm_model_name_or_path: dict = {
                ModelInterface.TYPE_TEXT: None,
                ModelInterface.TYPE_IMG: None,
            },
            fit_xform_model_name = 'cluster',
            cache_tensor_to_file = False,
            file_temp_dir = '/tmp',
            logger = None,
            ignore_warnings = False,
            return_key = False,
    ) -> VecDb:
        # If these params are different, they will return new object
        db_params = underlying_db_params
        key_id = str(VecDbClass) + '.lang=' + str(lang) + '.model_name=' + str(lm_model_name_or_path) \
                 + '.db_params=' + str(db_params.get_db_info())
        sgt = Singleton(
            class_type = VecDbClass,
            logger     = logger,
        ).get_singleton(
            key_id,
            user_id,
            lang,
            col_content,
            # text, image, sound, video, voice, etc.
            col_content_type,
            col_label_user,
            col_label_std,
            col_embedding,
            feature_len,
            underlying_db_params,
            # Where we cache the language models
            lm_cache_folder,
            # Language model name, e.g. "xlm-roberta-base", if empty will use default map by language
            lm_model_name_or_path,
            fit_xform_model_name,
            cache_tensor_to_file,
            file_temp_dir,
            logger,
            ignore_warnings,
        )
        return (sgt, key_id) if return_key else sgt


class VecDbSingletonUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        botid = 'mybot'
        vecdb = VecDbSingleton.get_singleton(
            VecDbClass = VecDb,
            user_id = botid,
            lang = 'multi',
            col_content = 'text',
            col_content_type = 'content_type',
            col_label_user = 'label',
            col_label_std = 'label_number',
            col_embedding = 'embedding',
            feature_len = 384,
            lm_cache_folder = Env().MODELS_PRETRAINED_DIR,
            # lm_model_name_or_path = None,
            fit_xform_model_name = os.environ["VECDB_FIT_XFORM_MODEL"],
            underlying_db_params = DbParams.get_db_params_from_envvars(
                identifier = 'test',
                db_create_tbl_sql = '',
                db_table = botid,
            ),
            cache_tensor_to_file = True,
            logger = self.logger,
        )
        vecdb.stop_threads()
        print('VECDB SINGLETON TESTS OK')
        return


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/env.ut.mysql')
    VecDbSingletonUnitTest().test()
    exit(0)



import logging
import os
import re
from fitxf import DatastoreInterface, DbParams
from fitxf.math.datasource.vecdb.model.ModelDb import ModelDb
from fitxf.math.datasource.vecdb.metadata.Metadata import ModelMetadata
from fitxf import ModelFitTransform
from fitxf import FitXformPca, FitXformCluster, FitXformClusterCosine
from fitxf.math.lang.encode.ImgPt import ImgPt
from fitxf.math.lang.encode.LangModelPtSingleton import LangModelInterface, LangModelPtSingleton
from fitxf.math.lang.encode.LangModelPt import LangModelPt
from fitxf.math.utils import Logging
from fitxf.math.utils.Env import Env


#
# Vector DBs are not yet mature, non-standard query languages, unable to do
# basic DB operations, knowledge in the field hard to find to maintain.
#
# We implement a fast & simple one here, for a single acc/bot id - as a unique
# DB table name.
#
class VecDb(DatastoreInterface):

    LangModelClass = LangModelPt

    MAX_RECORDS = 500000
    MAX_LEN_USER_ID = 64
    # "np" numpy arrays, "torch" Tensors also permitted
    RETURN_TENSORS = 'np'

    def __init__(
            self,
            user_id: str,
            lang: str,
            col_content: str,
            # text, image, sound, video, voice, etc.
            col_content_type: str,
            col_label_user: str,
            col_label_std: str,
            col_embedding: str,
            feature_len: int,
            underlying_db_params: DbParams,
            lm_cache_folder: str = None,
            # Hugging Face language model name, will use some default multilingual model if None
            lm_model_name_or_path: dict = {
                ModelFitTransform.TYPE_TEXT: None,
                ModelFitTransform.TYPE_IMG: None,
            },
            fit_xform_model_name: str = 'cluster',
            cache_tensor_to_file: bool = False,
            file_temp_dir: str = '/tmp',
            # Turn off fitting, thus in plain or test mode
            turn_off_fitting: bool = True,
            logger: Logging = None,
            ignore_warnings: bool = False,
    ):
        self.db_params = underlying_db_params
        if self.db_params.db_type == 'mysql':
            self.db_params.db_create_table_sql = \
                "CREATE TABLE `<TABLENAME>` (\n" + \
                "`" + str(col_content) + "` TEXT NOT NULL,\n" + \
                "`" + str(col_label_user) + "` varchar(255) NOT NULL,\n" + \
                "`" + str(col_label_std) + "` int NOT NULL,\n" + \
                "`" + str(col_embedding) + "` TEXT NOT NULL\n" + \
                ")"
            logger.info('Create table SQL for mysql: ' + str(self.db_params.db_create_table_sql))

        super().__init__(
            db_params = self.db_params,
            logger = logger,
            ignore_warnings = ignore_warnings,
        )
        self.user_id = user_id
        self.lang = lang
        self.col_content = col_content
        self.col_content_type = col_content_type
        self.col_label_user = col_label_user
        self.col_label_standardized = col_label_std
        self.col_embedding = col_embedding
        self.feature_len = feature_len
        self.llm_cache_folder = lm_cache_folder if lm_cache_folder is not None else Env.get_home_download_dir()
        self.llm_model_name_or_path = lm_model_name_or_path
        self.fit_xform_model_name = fit_xform_model_name

        if len(self.user_id) > self.MAX_LEN_USER_ID:
            raise Exception(
                'AccBot ID name too long > ' + str(self.MAX_LEN_USER_ID) + ': "' + str(self.user_id) + '"'
            )

        self.logger.info(
            'Before init, using index id "' + str(user_id) + '", lang "' + str(self.lang)
            + '", content column name "' + str(self.col_content) + '", content type "' + str(self.col_content_type)
            + '", user label "' + str(self.col_label_user)
            + '", label std "' + str(self.col_label_standardized) + '" embedding "' + str(self.col_embedding)
            + '", lm cache folder "' + str(self.llm_cache_folder)
            + '", lm model name or path "' + str(self.llm_model_name_or_path) + '"'
        )

        # Put large tensor objects in persistent object file
        self.cache_tensor_to_file = cache_tensor_to_file
        self.file_temp_dir = file_temp_dir
        self.turn_off_fitting = turn_off_fitting

        self.LangModelClass = {'text': LangModelPt, 'image': ImgPt}
        self.llm_model = {}
        for ctype in self.llm_model_name_or_path.keys():
            self.llm_model[ctype] = self.get_lang_model_class(
                lm_class = self.LangModelClass[ctype],
                llm_cache_folder = self.llm_cache_folder,
                llm_model_name_or_path = self.llm_model_name_or_path[ctype],
            )
            # Model name/path may have changed if we passed in None
            self.llm_model_name_or_path[ctype] = self.llm_model[ctype].get_model_name()
            self.logger.info(
                'Model type "' + str(ctype) + '" initialized with lang "' + str(self.lang)
                + '", lm cache folder "' + str(self.llm_cache_folder)
                + '", lm model name or path "' + str(self.llm_model_name_or_path)
                + '", cache tensor to file "' + str(self.cache_tensor_to_file)
                + '", temp dir "' + str(self.file_temp_dir) + '"'
            )

        self.fixed_tablename_or_index = self.user_id
        self.db_params.db_tablename = self.user_id

        self.logger.info(
            'Using content column name "' + str(self.col_content) + '", user label "' + str(self.col_label_user)
            + '", label std "' + str(self.col_label_standardized) + '" embedding "' + str(self.col_embedding)
            + '", fixed table/index "' + str(self.fixed_tablename_or_index)
            + '", underlying DB params: ' + str(self.db_params.get_db_info())
        )

        self.logger.info('Fit transform model name "' + str(self.fit_xform_model_name) + '"')
        if self.fit_xform_model_name == 'pca':
            self.FitClassObj = FitXformPca(logger=self.logger)
        elif self.fit_xform_model_name == 'cluster':
            self.FitClassObj = FitXformCluster(logger=self.logger)
        elif self.fit_xform_model_name == 'cluster-cosine':
            self.FitClassObj = FitXformClusterCosine(logger=self.logger)
        else:
            raise Exception('Unsupported model "' + str(self.fit_xform_model_name) + '"')

        self.logger.info('Fit transform model class "' + str(type(self.FitClassObj)) + '"')

        self.vec_db_model = ModelFitTransform(
            user_id = self.user_id,
            llm_model = self.llm_model,
            model_db_class = ModelDb,
            model_metadata_class = ModelMetadata,
            col_content = self.col_content,
            col_content_type = col_content_type,
            col_label_user = self.col_label_user,
            col_label_std = self.col_label_standardized,
            col_embedding = self.col_embedding,
            feature_len = self.feature_len,
            numpy_to_b64_for_db = True,
            fit_xform_model = self.FitClassObj,
            cache_tensor_to_file = self.cache_tensor_to_file,
            file_temp_dir = self.file_temp_dir,
            in_plain_or_test_mode = self.turn_off_fitting,
            return_tensors = self.RETURN_TENSORS,
            logger = self.logger,
        )
        self.logger.info(
            'ModelFitTransform initialized successfully for model type "' + str(ctype) + '"'
        )
        return

    def get_lang_model_class(
            self,
            lm_class: type(LangModelInterface),
            llm_cache_folder: str,
            llm_model_name_or_path: dict,
    ) -> LangModelInterface:
        return LangModelPtSingleton.get_singleton(
            LmClass = lm_class,
            cache_folder = llm_cache_folder,
            model_name = llm_model_name_or_path,
            logger = self.logger,
        )

    def get_metadata_table_name(
            self,
    ):
        # lower case to ensure compatibility with DBs such as Opensearch
        return 'metadata'

    def stop_threads(self):
        self.vec_db_model.stop_threads()
        return

    def get_data_length(self):
        return self.vec_db_model.get_data_length()

    # **
    def load_data_model(
            self,
            max_tries = 1,
            background = False,
    ):
        return self.vec_db_model.load_data_model(
            max_tries = max_tries,
            background = background,
        )

    def connect(
            self,
            host     = None,
            port     = None,
            username = None,
            password = None,
            database = None,
            scheme   = None,
            # For our Soprano network, this must be False, otherwise too many problems with CA Authority
            verify_certs = True,
            other_params = None,
    ):
        # already connected upon __init__(), nothing to do
        return

    def get(
            self,
            # e.g. {"answer": "take_seat"}
            match_phrase,
            tablename = None,
            request_timeout = 20,
            params_other = None,
    ):
        if tablename is not None:
            assert tablename == self.fixed_tablename_or_index, str(tablename) + ' != ' + str(self.fixed_tablename_or_index)
        return self.vec_db_model.model_db.underlying_db.get(
            match_phrase = match_phrase,
            tablename = tablename,
            request_timeout = request_timeout,
        )

    def get_all(
            self,
            key = None,
            max_records = 10000,
            tablename_or_index = None,
            return_db_style_records  =True,
            request_timeout = 20,
            params_other = None,
    ):
        if tablename_or_index is not None:
            assert tablename_or_index == self.fixed_tablename_or_index, str(tablename_or_index) + ' != ' + str(self.fixed_tablename_or_index)
        return self.vec_db_model.model_db.underlying_db.get_all(
            key = key,
            max_records = max_records,
            tablename = tablename_or_index,
            request_timeout = request_timeout,
        )

    def get_indexes(self):
        return self.vec_db_model.model_db.underlying_db.get_indexes()

    def delete_index(
            self,
            tablename_or_index,
    ):
        if tablename_or_index is not None:
            assert tablename_or_index == self.fixed_tablename_or_index, str(tablename_or_index) + ' != ' + str(self.fixed_tablename_or_index)
        return self.vec_db_model.model_db.underlying_db.delete_index(
            tablename = tablename_or_index,
        )

    def atomic_delete_add(
            self,
            delete_key: str,
            # list of dicts
            records: list[dict],
            model: str = None,
            params_other = None,
    ):
        return self.vec_db_model.atomic_delete_add(
            delete_key = delete_key,
            records = records,
        )

    def add(
            self,
            # list of dicts
            records,
            model: str = None,
            params_other = None,
    ):
        return self.vec_db_model.add(
            records = records,
        )

    def delete(
            self,
            match_phrase: list[dict],
            tablename_or_index: str = None,
            params_other = None,
    ):
        if tablename_or_index is not None:
            assert tablename_or_index == self.fixed_tablename_or_index, str(tablename_or_index) + ' != ' + str(self.fixed_tablename_or_index)
        match_phrases = match_phrase if type(match_phrase) in [list, tuple] else [match_phrase]

        try:
            return self.vec_db_model.delete(
                match_phrases = match_phrases,
            )
        except Exception as ex:
            self.logger.warning(
                'Error delete from table "' + str(tablename_or_index) + '" match phrases ' + str(match_phrases)
            )

    def cleanup_underlying_db(
            self,
    ):
        try:
            self.vec_db_model.model_db.underlying_db.delete_index(tablename=self.fixed_tablename_or_index)
        except Exception as ex:
            self.logger.error('Error delete DB index "' + str(self.fixed_tablename_or_index) + '": ' + str(ex))
        res = self.vec_db_model.vec_db_metadata.cleanup()
        self.logger.info('Delete result ' + str(res))
        return

    def predict(
            self,
            text_list_or_embeddings,
            content_type = ModelFitTransform.TYPE_TEXT,
            top_k = 5,
            # Instead of just returning the user labels, return full record. Applicable to some models only
            return_full_record = False,
    ):
        return self.vec_db_model.predict(
            text_list_or_embeddings = text_list_or_embeddings,
            content_type = content_type,
            top_k = top_k,
            return_full_record = return_full_record,
        )


if __name__ == '__main__':
    er = Env()
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.nwae.math.ut')
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    col_content = 'data'
    col_type = 'type'
    col_label_user = 'label'
    col_label_std = 'label_n'
    col_embedding = 'enc'
    vdb = VecDb(
        user_id = 'main_test_vecdb',
        lang = 'en',
        col_content = col_content,
        col_content_type = col_type,
        col_label_user = col_label_user,
        col_label_std = col_label_std,
        col_embedding = col_embedding,
        feature_len = 768,
        lm_cache_folder = er.MODELS_PRETRAINED_DIR,
        fit_xform_model_name = 'cluster',
        # The other info will be auto filled from env var
        underlying_db_params = DbParams(identifier='testvecdb', logger=lgr),
        logger = lgr,
    )
    res = vdb.add(
        records = [
            {col_label_user: 'hola', col_type: 'text', col_content: 'hi'},
            {col_label_user: 'hola', col_type: 'text', col_content: 'hola'},
            {col_label_user: 'hola', col_type: 'text', col_content: 'ohayo'},
            {col_label_user: 'test', col_type: 'text', col_content: 'testing'},
            {col_label_user: 'test', col_type: 'text', col_content: 'cqcq'},
            {col_label_user: 'test', col_type: 'text', col_content: '试试'},
            {col_label_user: 'plov', col_type: 'image', col_content: 'https://img.freepik.com/premium-photo/shakh-plov-cooked-rice-dish-with-raisins-beautiful-plate-islamic-arabic-food_1279579-5074.jpg?w=1800'},
            {col_label_user: 'plov', col_type: 'image', col_content: 'https://img.freepik.com/premium-psd/tasty-fried-vegetable-rice-plate-isolated-transparent-background_927015-3126.jpg?w=1480'},
        ],
    )
    print('add result', res)
    for txt in [
        '안녕', 'привет', '시험하다', 'проверь',
        'https://img.freepik.com/premium-psd/tasty-fried-vegetable-rice-plate-isolated-transparent-background_927015-3126.jpg?w=1480',
    ]:
        content_type = ModelFitTransform.TYPE_IMG if re.match(pattern="^http", string=txt) \
            else ModelFitTransform.TYPE_TEXT
        matches, probs = vdb.predict(
            text_list_or_embeddings = [txt],
            content_type = content_type,
        )
        print('predict result for "' + str(txt) + '"')
        [print(r) for r in list(zip(matches[0], probs[0]))]

    vdb.stop_threads()
    exit(0)
